/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling.model;

import org.teleal.common.util.URIUtil;

import java.net.InetAddress;
import java.net.URI;
import java.net.URL;
import java.util.Arrays;

/**
 * The URL and MAC address a device (descriptor) is reachable at.
 * <p>
 * Used when sending notification messages about local devices and services to
 * other UPnP participants on the network. We send them the URL our descriptors
 * are reachable at, and we also let them know our MAC hardware address so they
 * can wake us up from sleep with Wake-On-LAN if necessary.
 * </p>
 *
 * @author Christian Bauer
 */
public class Locator {

    protected InetAddress address;
    protected int port;
    protected URI path;
    protected byte[] hardwareAddress;

    public Locator(InetAddress address, int port, URI path) {
        this(address, port, path, null);
    }

    public Locator(InetAddress address, int port, URI path, byte[] hardwareAddress) {
        this.address = address;
        this.port = port;
        this.path = path;
        this.hardwareAddress = hardwareAddress;
    }

    public InetAddress getAddress() {
        return address;
    }

    public int getPort() {
        return port;
    }

    public URI getPath() {
        return path;
    }

    public byte[] getHardwareAddress() {
        return hardwareAddress;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        Locator locator = (Locator) o;

        if (port != locator.port) return false;
        if (!address.equals(locator.address)) return false;
        if (!Arrays.equals(hardwareAddress, locator.hardwareAddress)) return false;
        if (!path.equals(locator.path)) return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = address.hashCode();
        result = 31 * result + port;
        result = 31 * result + path.hashCode();
        result = 31 * result + (hardwareAddress != null ? Arrays.hashCode(hardwareAddress) : 0);
        return result;
    }

    /**
     * @return An HTTP URL with the address, port, and path of this locator.
     */
    public URL getURL() {
        return URIUtil.createAbsoluteURL(address, port, getPath());
    }

}
