/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling.test.support.connectionmanager;

import org.teleal.cling.binding.annotations.AnnotationLocalServiceBinder;
import org.teleal.cling.model.DefaultServiceManager;
import org.teleal.cling.model.ServiceReference;
import org.teleal.cling.model.ValidationException;
import org.teleal.cling.model.action.ActionException;
import org.teleal.cling.model.action.ActionInvocation;
import org.teleal.cling.model.message.UpnpResponse;
import org.teleal.cling.model.meta.DeviceDetails;
import org.teleal.cling.model.meta.DeviceIdentity;
import org.teleal.cling.model.meta.LocalDevice;
import org.teleal.cling.model.meta.LocalService;
import org.teleal.cling.model.types.UDADeviceType;
import org.teleal.cling.model.types.UDN;
import org.teleal.cling.support.connectionmanager.AbstractPeeringConnectionManagerService;
import org.teleal.cling.support.connectionmanager.dto.ConnectionInfo;
import org.teleal.cling.support.connectionmanager.dto.ProtocolInfo;
import org.teleal.cling.support.connectionmanager.dto.ProtocolInfos;
import org.testng.annotations.Test;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import static org.testng.Assert.assertEquals;


/**
 * @author Christian Bauer
 */
public class ConnectionManagerPeerTest {

    @Test
    public void createDestroyConnections() throws Exception {

        CountingListener listener = new CountingListener();

        PeeringConnectionManager mediaServer =
                new PeeringConnectionManager(
                        new ProtocolInfos("http-get:*:video/mpeg:*,http-get:*:audio/mpeg:*"),
                        null
                );
        mediaServer.getPropertyChangeSupport().addPropertyChangeListener(listener);
        LocalService<PeeringConnectionManager> mediaServerMeta = createService(mediaServer);
        createDevice("MEDIASERVER-AAA-AAA-AAA", "MediaServer", mediaServerMeta);

        // The sink
        PeeringConnectionManager mediaRenderer =
                new PeeringConnectionManager(
                        null,
                        new ProtocolInfos("http-get:*:video/mpeg:*")
                );
        mediaRenderer.getPropertyChangeSupport().addPropertyChangeListener(listener);
        LocalService<PeeringConnectionManager> mediaRendererMeta = createService(mediaRenderer);
        createDevice("MEDIARENDERER-BBB-BBB-BBB", "MediaRenderer", mediaRendererMeta);

        // Connect MediaRenderer for input with MediaServer output
        mediaServer.createConnectionWithPeer(
                mediaServerMeta.getReference(),
                null,
                mediaRendererMeta,
                new ProtocolInfo("http-get:*:video/mpeg:*"),
                ConnectionInfo.Direction.Input
        );

        assertEquals(mediaServer.getCurrentConnectionIDs().size(), 1);
        assertEquals(mediaRenderer.getCurrentConnectionIDs().size(), 1);

        assertEquals(mediaRenderer.getCurrentConnectionInfo(0).getConnectionID(), mediaServer.getCurrentConnectionInfo(0).getPeerConnectionID());

        assertEquals(mediaServer.getCurrentConnectionInfo(0).getDirection(), ConnectionInfo.Direction.Output);
        assertEquals(mediaRenderer.getCurrentConnectionInfo(0).getDirection(), ConnectionInfo.Direction.Input);

        assertEquals(mediaServer.getCurrentConnectionInfo(0).getRcsID(), 111);
        assertEquals(mediaRenderer.getCurrentConnectionInfo(0).getRcsID(), 111);
        assertEquals(mediaServer.getCurrentConnectionInfo(0).getAvTransportID(), 333);
        assertEquals(mediaRenderer.getCurrentConnectionInfo(0).getAvTransportID(), 333);

        assertEquals(mediaServer.getCurrentConnectionInfo(0).getConnectionStatus(), ConnectionInfo.Status.OK);
        assertEquals(mediaRenderer.getCurrentConnectionInfo(0).getConnectionStatus(), ConnectionInfo.Status.OK);

        // Another connection
        mediaServer.createConnectionWithPeer(
                mediaServerMeta.getReference(),
                null,
                mediaRendererMeta,
                new ProtocolInfo("http-get:*:video/mpeg:*"),
                ConnectionInfo.Direction.Input
        );

        assertEquals(mediaServer.getCurrentConnectionIDs().size(), 2);
        assertEquals(mediaRenderer.getCurrentConnectionIDs().size(), 2);

        assertEquals(mediaServer.getCurrentConnectionInfo(1).getRcsID(), 222);
        assertEquals(mediaRenderer.getCurrentConnectionInfo(1).getRcsID(), 222);
        assertEquals(mediaServer.getCurrentConnectionInfo(1).getAvTransportID(), 444);
        assertEquals(mediaRenderer.getCurrentConnectionInfo(1).getAvTransportID(), 444);

        // Close one
        mediaServer.closeConnectionWithPeer(null, mediaRendererMeta, mediaServer.getCurrentConnectionInfo(0));

        assertEquals(mediaServer.getCurrentConnectionIDs().size(), 1);
        assertEquals(mediaRenderer.getCurrentConnectionIDs().size(), 1);

        // The other is still there
        assertEquals(mediaServer.getCurrentConnectionInfo(1).getPeerConnectionID(), 1);

        // Should have 2 + 2 + 2 (connect, connect, disconnect on both connectionmanagers) events
        assertEquals(listener.count, 6);
    }

    public LocalDevice createDevice(String udn, String type, LocalService service) throws ValidationException {
        return new LocalDevice(
                new DeviceIdentity(new UDN(udn)),
                new UDADeviceType(type),
                new DeviceDetails(type),
                service
        );
    }

    public LocalService<PeeringConnectionManager> createService(final PeeringConnectionManager peer) {
        LocalService<PeeringConnectionManager> service =
                new AnnotationLocalServiceBinder().read(AbstractPeeringConnectionManagerService.class);
        service.setManager(
                new DefaultServiceManager<PeeringConnectionManager>(service, null) {
                    @Override
                    protected PeeringConnectionManager createServiceInstance() throws Exception {
                        return peer;
                    }

                    // This just helps debugging of this test
                    @Override
                    public String toString() {
                        return getService().getReference().toString();
                    }
                }
        );
        return service;
    }

    class PeeringConnectionManager extends AbstractPeeringConnectionManagerService {

        PeeringConnectionManager(ProtocolInfos sourceProtocolInfo, ProtocolInfos sinkProtocolInfo, ConnectionInfo... activeConnections) {
            super(sourceProtocolInfo, sinkProtocolInfo, activeConnections);
        }

        @Override
        protected ConnectionInfo createConnection(int connectionID, int peerConnectionId, ServiceReference peerConnectionManager,
                                                  ConnectionInfo.Direction direction, ProtocolInfo protocolInfo) throws ActionException {

            // Some rcs/av transport IDs for testing
            return new ConnectionInfo(
                    connectionID,
                    connectionID == 0 ? 111 : 222,
                    connectionID == 0 ? 333 : 444,
                    protocolInfo,
                    peerConnectionManager,
                    peerConnectionId,
                    direction,
                    ConnectionInfo.Status.OK
            );
        }

        @Override
        protected void closeConnection(ConnectionInfo connectionInfo) {
            // NOOP
        }

        @Override
        protected void peerFailure(ActionInvocation invocation, UpnpResponse operation, String defaultFailureMessage) {
            System.err.println("Error managing connection with peer: " + defaultFailureMessage);
        }
    }

    class CountingListener implements PropertyChangeListener {
        int count = 0;

        public void propertyChange(PropertyChangeEvent e) {
            if (e.getPropertyName().equals("CurrentConnectionIDs")) {
                count++;
            }
        }
    }

}
