/*
 * Copyright (C) 2010 Teleal GmbH, Switzerland
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.teleal.cling.test.support.contentdirectory;

import org.teleal.cling.binding.annotations.AnnotationLocalServiceBinder;
import org.teleal.cling.model.DefaultServiceManager;
import org.teleal.cling.model.action.ActionInvocation;
import org.teleal.cling.model.meta.LocalService;
import org.teleal.cling.model.meta.Service;
import org.teleal.cling.support.contentdirectory.AbstractContentDirectoryService;
import org.teleal.cling.support.contentdirectory.BrowseActionCallback;
import org.teleal.cling.support.contentdirectory.BrowseFlag;
import org.teleal.cling.support.contentdirectory.BrowseResult;
import org.teleal.cling.support.contentdirectory.ContentDirectoryErrorCode;
import org.teleal.cling.support.contentdirectory.ContentDirectoryException;
import org.teleal.cling.support.contentdirectory.SortCriterion;
import org.teleal.cling.support.contentdirectory.descriptor.DIDLParser;
import org.teleal.cling.support.contentdirectory.descriptor.didl.Container;
import org.teleal.cling.support.contentdirectory.descriptor.didl.DIDLLite;
import org.testng.annotations.Test;

import java.io.StringWriter;

import static org.testng.Assert.assertEquals;

/**
 * @author Christian Bauer
 */
public class ContentDirectoryBrowseTest {

    protected DIDLParser parser = new DIDLParser();

    @Test
    public void browseRootMetadata() {

        final boolean[] assertions = new boolean[3];
        new BrowseActionCallback(createService(), "0", BrowseFlag.METADATA) {
            @Override
            public void received(ActionInvocation actionInvocation, DIDLLite didl) {
                assertEquals(didl.getChildren().size(), 1);
                assertEquals(((Container)didl.getChildren().get(0)).getTitle().getValue(), "My multimedia stuff");
                assertions[0] = true;
            }

            @Override
            public void updateStatus(Status status) {
                if (!assertions[1] && status.equals(Status.LOADING)) {
                    assertions[1] = true;
                } else if (assertions[1] && !assertions[2] && status.equals(Status.OK)) {
                    assertions[2] = true;
                }
            }

            @Override
            public void failure(String failureMessage) {
            }
        }.run();

        for (boolean assertion : assertions) {
            assertEquals(assertion, true);
        }
    }

    @Test
    public void browseRootChildren() {

        final boolean[] assertions = new boolean[3];
        new BrowseActionCallback(
                createService(), "0", BrowseFlag.DIRECT_CHILDREN, "foo", 1, 10l,
                new SortCriterion(true, "dc:title"), new SortCriterion(false, "dc:creator")
        ) {
            @Override
            public void received(ActionInvocation actionInvocation, DIDLLite didl) {
                assertEquals(didl.getChildren().size(), 3);
                assertEquals(((Container)didl.getChildren().get(0)).getTitle().getValue(), "My Music");
                assertEquals(((Container)didl.getChildren().get(1)).getTitle().getValue(), "My Photos");
                assertEquals(((Container)didl.getChildren().get(2)).getTitle().getValue(), "Album Art");
                assertions[0] = true;
            }

            @Override
            public void updateStatus(Status status) {
                if (!assertions[1] && status.equals(Status.LOADING)) {
                    assertions[1] = true;
                } else if (assertions[1] && !assertions[2] && status.equals(Status.OK)) {
                    assertions[2] = true;
                }
            }

            @Override
            public void failure(String failureMessage) {
            }
        }.run();

        for (boolean assertion : assertions) {
            assertEquals(assertion, true);
        }
    }

    public Service createService() {
        LocalService<AbstractContentDirectoryService> service =
                new AnnotationLocalServiceBinder().read(AbstractContentDirectoryService.class);
        service.setManager(
                new DefaultServiceManager<AbstractContentDirectoryService>(service, null) {
                    @Override
                    protected AbstractContentDirectoryService createServiceInstance() throws Exception {
                        return createCDS();
                    }
                }
        );
        return service;
    }

    public AbstractContentDirectoryService createCDS() {
        return new AbstractContentDirectoryService() {

            @Override
            public BrowseResult browse(String objectID, BrowseFlag browseFlag, String filter,
                                       long firstResult, long maxResults, SortCriterion[] orderby) throws ContentDirectoryException {

                try {

                    if (objectID.equals("0") && browseFlag.equals(BrowseFlag.METADATA)) {

                        assertEquals(firstResult, 0);
                        assertEquals(maxResults, 999);

                        assertEquals(orderby.length, 0);

                        String result = readDIDLResource("org/teleal/cling/test/support/contentdirectory/samples/browseRoot.xml");
                        return new BrowseResult(result, 1, 1);

                    } else if (objectID.equals("0") && browseFlag.equals(BrowseFlag.DIRECT_CHILDREN)) {

                        assertEquals(filter, "foo");
                        assertEquals(firstResult, 1);
                        assertEquals(maxResults, 10l);

                        assertEquals(orderby.length, 2); // We don't sort, just test
                        assertEquals(orderby[0].isAscending(), true);
                        assertEquals(orderby[0].getPropertyName(), "dc:title");
                        assertEquals(orderby[1].isAscending(), false);
                        assertEquals(orderby[1].getPropertyName(), "dc:creator");

                        String result = readDIDLResource("org/teleal/cling/test/support/contentdirectory/samples/browseRootChildren.xml");
                        return new BrowseResult(result, 3, 3);
                    }

                    DIDLLite didl = new DIDLLite();
                    StringWriter writer = new StringWriter();
                    parser.createMarshaller().marshal(didl, writer);
                    return new BrowseResult(writer.toString(), 0, 0);

                } catch (Exception ex) {
                    throw new ContentDirectoryException(ContentDirectoryErrorCode.CANNOT_PROCESS, ex.toString());
                }
            }

        };
    }

    protected String readDIDLResource(String path) {
        try {
            DIDLLite didl = parser.readResource(path);
            StringWriter writer = new StringWriter(100);
            parser.createMarshaller().marshal(didl, writer);
            return writer.toString();
        } catch (Exception ex) {
            throw new RuntimeException(ex);
        }
    }

}
